r""":mod:`prolint2.plotting.multiples`
==========================================================
:Authors: Daniel P. Ramirez & Besian I. Sejdiu
:Year: 2022
:Copyright: MIT License
"""

import os
import math
import pandas as pd
import numpy as np
import networkx as nx
import logomaker as lm
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.patches import FancyBboxPatch
from matplotlib.ticker import FormatStrFormatter
from prolint2.computers.distances import SerialDistances
from prolint2.computers.distances import TwoPointDistances
from matplotlib.cm import ScalarMappable
import inspect
from scipy import interpolate
from prolint2.server.chord_utils import contact_chord
from .utils import *
from .plotting import Plotter

## seaborn config for paper quality plots
import seaborn as sns

sns.set_context("paper")
sns.set_style("whitegrid")
sns.set_palette("colorblind")

__all__ = [
    "MultiplePointDistribution",
    "MultipleRadar",
    "MultipleMosaics",
]


class MultiplePointDistribution(Plotter):
    """
    Initialize the MultiplePointDistribution instance.

    Parameters
    ----------
    universe : Universe
        The universe containing the data for the plot.
    metric : dict
        A dictionary containing the metric data.
    xlabel : str, optional
        The label for the x-axis. Default is None.
    ylabel : str, optional
        The label for the y-axis. Default is None.
    fn : str, optional
        The filename to save the plot. Default is None.
    title : str, optional
        The title for the plot. Default is None.
    fig_size : tuple, optional
        The figure size as a tuple (width, height). Default is (8, 8).

    Returns
    -------
    None
    """

    def __init__(
        self,
        universe,
        metric,
        xlabel=None,
        ylabel=None,
        fn=None,
        title=None,
        fig_size=(8, 8),
    ):
        # Initialize the PointDistribution instance.
        # Inherits from Plotter and sets plot labels, title, and figure size.
        super().__init__(xlabel, ylabel, fn, title, fig_size)
        self.universe = universe
        self.metric = metric

    def _plot_single(self, lipid_type, metric_name, axis, res_ids=None, **kwargs):
        """
        Plot a scatter plot for a specific lipid type and metric.

        Parameters
        ----------
        lipid_type : str
            The lipid type to plot.
        metric_name : str
            The name of the metric to plot.
        axis : matplotlib axis
            The axis for the scatter plot.
        res_ids : list, optional
            List of residue IDs to include. Default is None.
        **kwargs
            Additional keyword arguments for the scatterplot.

        Returns
        -------
        axis : matplotlib axis
            The updated axis containing the scatter plot.
        """
        # Get the list of metric values for specified lipid_type and metric_name
        res_list, metric_list = get_metric_list_by_residues(
            self.universe, self.metric, lipid_type, metric_name, res_list=res_ids
        )

        # Create scatter plot without palette
        axis = sns.scatterplot(x=res_list, y=metric_list, ax=axis, **kwargs)

        axis.tick_params(axis="both", which="major", labelsize=10)
        axis.set_title(
            "{} | {}".format(lipid_type, metric_name),
            fontsize=11,
            fontfamily="Arial Unicode MS",
        )
        # add axes labels
        axis.set_xlabel(self.xlabel, fontsize=12, fontfamily="Arial Unicode MS")

        return axis

    def create_plot(self, rows="metrics", res_ids=None, decimals=1, **kwargs):
        """
        Create the multiple point distribution plot.

        Parameters
        ----------
        rows : str
            Specify whether to arrange plots by 'metrics' or 'lipids'.
        res_ids : list, optional
            List of residue IDs to include in the plot. Default is None.
        decimals : int, optional
            Number of decimal places to format y-axis labels. Default is 1.
        **kwargs
            Additional keyword arguments for the plot.

        Returns
        -------
        None
        """
        if rows not in ["metrics", "lipids"]:
            raise ValueError("rows must be either 'metrics' or 'lipids'.")

        n_lipids = len(self.universe.database.unique_resnames)
        n_metrics = len(
            self.metric[list(self.metric.keys())[0]][
                list(self.metric[list(self.metric.keys())[0]].keys())[0]
            ]
        )
        if n_lipids == 1 and n_metrics == 1:
            raise ValueError(
                "You only have one lipid type and one metric. We encourage you to use a single plot instead."
            )

        metric_names = []
        # Collect all unique metric names from the metric dictionary
        for res in self.metric:
            for lip in self.metric[res]:
                for key in self.metric[res][lip]:
                    if key not in metric_names:
                        metric_names.append(key)

        if rows == "metrics":
            n_rows = n_metrics
            n_cols = n_lipids
        else:
            n_rows = n_lipids
            n_cols = n_metrics

        # set default labels
        if self.xlabel is None:
            self.xlabel = "Residue ID"
        if self.ylabel is None:
            self.ylabel = "Metric Value"

        if rows == "metrics":
            fig, axes = plt.subplots(
                n_rows, n_cols, figsize=self.fig_size, sharex=True, sharey="row"
            )
            if n_rows == 1 and n_cols > 1:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    axes[i] = self._plot_single(
                        lipid_type, metric_names[0], axes[i], res_ids=res_ids, **kwargs
                    )
            elif n_rows > 1 and n_cols == 1:
                for i, metric_name in enumerate(metric_names):
                    axes[i] = self._plot_single(
                        self.universe.database.unique_resnames[0],
                        metric_name,
                        axes[i],
                        res_ids=res_ids,
                        **kwargs
                    )
            else:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    for j, metric_name in enumerate(metric_names):
                        axes[j, i] = self._plot_single(
                            lipid_type,
                            metric_name,
                            axes[j, i],
                            res_ids=res_ids,
                            **kwargs
                        )
        elif rows == "lipids":
            fig, axes = plt.subplots(
                n_rows, n_cols, figsize=self.fig_size, sharex=True, sharey="col"
            )
            if n_rows == 1 and n_cols > 1:
                for i, metric_name in enumerate(metric_names):
                    axes[i] = self._plot_single(
                        self.universe.database.unique_resnames[0],
                        metric_name,
                        axes[i],
                        res_ids=res_ids,
                        **kwargs
                    )
            elif n_rows > 1 and n_cols == 1:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    axes[i] = self._plot_single(
                        lipid_type, metric_names[0], axes[i], res_ids=res_ids, **kwargs
                    )
            else:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    for j, metric_name in enumerate(metric_names):
                        axes[i, j] = self._plot_single(
                            lipid_type,
                            metric_name,
                            axes[i, j],
                            res_ids=res_ids,
                            **kwargs
                        )

        # format decimal places in yaxis labels
        for ax in axes.flat:
            ax.yaxis.set_major_formatter(FormatStrFormatter("%.{}f".format(decimals)))

        # adding ylabel to the first column
        if n_rows == 1:
            axes[0].set_ylabel(self.ylabel, fontsize=12, fontfamily="Arial Unicode MS")
        else:
            for ax in axes[:, 0]:
                ax.set_ylabel(self.ylabel, fontsize=12, fontfamily="Arial Unicode MS")

        # Set default filename, and title if not provided
        if self.fn is None:
            self.fn = os.path.join(
                os.getcwd(),
                "multiple_point_distribution.pdf",
            )
        if self.title is None:
            if rows == "metrics":
                self.title = "Point distributions - Metrics x Lipid types"
            else:
                self.title = "Point distributions - Lipid types x Metrics"

        # # Add title to the plot
        fig.suptitle(
            self.title,
            fontsize=14,
            weight="bold",
            fontfamily="Arial Rounded MT Bold",
        )
        plt.tight_layout()


class MultipleRadar(Plotter):
    """
    Initialize a MultipleRadar instance.

    Parameters
    ----------
    universe : Universe
        The molecular dynamics universe.
    metric : dict
        A dictionary containing metrics for lipids.
    xlabel : str, optional
        The label for the x-axis.
    ylabel : str, optional
        The label for the y-axis.
    fn : str, optional
        The filename to save the plot.
    title : str, optional
        The title of the radar chart.
    fig_size : tuple, optional
        The figure size in inches (width, height).

    Inherits from Plotter and sets plot labels, title, and figure size.
    """

    def __init__(
        self,
        universe,
        metric,
        xlabel=None,
        ylabel=None,
        fn=None,
        title=None,
        fig_size=(8, 8),
    ):
        # Initialize the PointDistribution instance.
        # Inherits from Plotter and sets plot labels, title, and figure size.
        super().__init__(xlabel, ylabel, fn, title, fig_size)
        self.universe = universe
        self.metric = metric

    def _plot_single(
        self, lipid_type, metric_name, axis, theta, res_ids=None, **kwargs
    ):
        """
        Plot a radar chart for a single lipid type and metric.

        Parameters
        ----------
        lipid_type : str
            The lipid type.
        metric_name : str
            The metric name.
        axis : matplotlib.axes._subplots.PolarAxesSubplot
            The polar axes for plotting.
        theta : numpy.ndarray
            The angles for the radar chart.
        res_ids : list, optional
            A list of residue IDs to include in the plot.
        kwargs : dict
            Additional keyword arguments for plotting.

        Returns
        -------
        matplotlib.axes._subplots.PolarAxesSubplot
            The polar axes with the radar chart.
        """
        # Get a list of metrics for the given lipid and metric_name
        res_list, metric_list = get_metric_list_by_residues(
            self.universe, self.metric, lipid_type, metric_name, res_list=res_ids
        )

        axis.set_theta_zero_location("S")
        axis.set_theta_direction(-1)

        # Plot the radar chart bars
        axis.bar(theta, metric_list, width=0.02, alpha=0.5, **kwargs)

        axis.set_xlim(0, 2 * np.pi)
        axis.xaxis.grid(True, linestyle="dashed", alpha=0.5)
        axis.set_yticks([max(metric_list) / 2, max(metric_list)])
        axis.set_yticklabels([])

        axis.tick_params(axis="x", pad=10, labelsize=8)
        axis.yaxis.grid(True, linestyle="dashed", alpha=0.5)
        axis.set_rlabel_position(0)
        axis.set_thetamin(15)
        axis.set_thetamax(345)

        axis.set_title(
            "{} | {}".format(lipid_type, metric_name),
            fontsize=11,
            pad=10,
            fontfamily="Arial Rounded MT Bold",
        )

        return axis

    def create_plot(
        self, rows="metrics", res_ids=None, decimals=1, magic_number=None, **kwargs
    ):
        """
        Create a radar chart plot.

        Parameters
        ----------
        rows : str, optional
            Determines the layout of the plot. Must be 'metrics' or 'lipids'.
        res_ids : list, optional
            A list of residue IDs to include in the plot.
        decimals : int, optional
            The number of decimal places for labels.
        magic_number : int, optional
            A factor to determine the number of variables to display.
        kwargs : dict
            Additional keyword arguments for plotting.

        Raises
        ------
        ValueError
            If rows is neither 'metrics' nor 'lipids'.

        """
        if rows not in ["metrics", "lipids"]:
            raise ValueError("rows must be either 'metrics' or 'lipids'.")

        n_lipids = len(self.universe.database.unique_resnames)
        n_metrics = len(
            self.metric[list(self.metric.keys())[0]][
                list(self.metric[list(self.metric.keys())[0]].keys())[0]
            ]
        )
        if n_lipids == 1 and n_metrics == 1:
            raise ValueError(
                "You only have one lipid type and one metric. We encourage you to use a single plot instead."
            )

        metric_names = []
        # Collect all unique metric names from the metric dictionary
        for res in self.metric:
            for lip in self.metric[res]:
                for key in self.metric[res][lip]:
                    if key not in metric_names:
                        metric_names.append(key)

        if rows == "metrics":
            n_rows = n_metrics
            n_cols = n_lipids
        else:
            n_rows = n_lipids
            n_cols = n_metrics

        resids_total_list = self.universe.query.residues.resids.tolist()
        if res_ids is None:
            res_list = resids_total_list
        else:
            res_list = res_ids

        resnames = [
            self.universe.query.residues.resnames[j]
            for j in [resids_total_list.index(i) for i in res_list]
        ]

        # Calculate the number of variables and the corresponding angles for the radar chart
        num_vars = len(res_list)
        theta = np.linspace(0, 2 * np.pi - np.pi / 6, num_vars, endpoint=False)
        theta += np.pi / 12

        if magic_number is None:
            magic_number = len(res_list) // 15 + 1

        if rows == "metrics":
            fig, axes = plt.subplots(
                n_rows,
                n_cols,
                figsize=self.fig_size,
                sharey="row",
                subplot_kw={"polar": True},
            )
            if n_rows == 1 and n_cols > 1:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    axes[i] = self._plot_single(
                        lipid_type,
                        metric_names[0],
                        axes[i],
                        theta,
                        res_ids=res_ids,
                        **kwargs
                    )
            elif n_rows > 1 and n_cols == 1:
                for i, metric_name in enumerate(metric_names):
                    axes[i] = self._plot_single(
                        self.universe.database.unique_resnames[0],
                        metric_name,
                        axes[i],
                        theta,
                        res_ids=res_ids,
                        **kwargs
                    )
            else:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    for j, metric_name in enumerate(metric_names):
                        axes[j, i] = self._plot_single(
                            lipid_type,
                            metric_name,
                            axes[j, i],
                            theta,
                            res_ids=res_ids,
                            **kwargs
                        )
        elif rows == "lipids":
            fig, axes = plt.subplots(
                n_rows,
                n_cols,
                figsize=self.fig_size,
                sharey="col",
                subplot_kw={"polar": True},
            )
            if n_rows == 1 and n_cols > 1:
                for i, metric_name in enumerate(metric_names):
                    axes[i] = self._plot_single(
                        self.universe.database.unique_resnames[0],
                        metric_name,
                        axes[i],
                        theta,
                        res_ids=res_ids,
                        **kwargs
                    )
            elif n_rows > 1 and n_cols == 1:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    axes[i] = self._plot_single(
                        lipid_type,
                        metric_names[0],
                        axes[i],
                        theta,
                        res_ids=res_ids,
                        **kwargs
                    )
            else:
                for i, lipid_type in enumerate(self.universe.database.unique_resnames):
                    for j, metric_name in enumerate(metric_names):
                        axes[i, j] = self._plot_single(
                            lipid_type,
                            metric_name,
                            axes[i, j],
                            theta,
                            res_ids=res_ids,
                            **kwargs
                        )

        for ax in axes.flat:
            ax.set_xticks(theta[::magic_number])
            ax.set_xticklabels(
                [
                    "{} {}".format(
                        resnames[i],
                        res_list[i],
                    )
                    for i in range(0, len(res_list), magic_number)
                ],
                fontfamily="Arial Unicode MS",
            )

        for ax in axes.flat:
            yticks = ax.get_yticks()
            for loc in yticks:
                ax.text(
                    0.0,
                    loc,
                    str(round(float(loc), decimals)),
                    ha="center",
                    va="top",
                    fontsize=8,
                    fontfamily="Arial Unicode MS",
                )

        # Set default filename, and title if not provided
        if self.fn is None:
            self.fn = os.path.join(
                os.getcwd(),
                "multiple_radars.pdf",
            )
        if self.title is None:
            if rows == "metrics":
                self.title = "Radar - Metrics x Lipid types"
            else:
                self.title = "Radar - Lipid types x Metrics"

        # # Add title to the plot
        fig.suptitle(
            self.title,
            fontsize=14,
            weight="bold",
            fontfamily="Arial Rounded MT Bold",
        )
        plt.tight_layout()


class MultipleMosaics(Plotter):
    """
    Initialize the MultipleMosaics object with specified plot attributes.

    Parameters
    ----------
    list_of_grids : list of str
        List of filenames containing grid data.
    xlabel : str, optional
        Label for the x-axis of the plot.
    ylabel : str, optional
        Label for the y-axis of the plot.
    fn : str, optional
        Filename to save the plot. If None, the plot won't be saved.
    title : str, optional
        Title for the plot.
    fig_size : tuple, optional
        Figure size in inches (width, height).

    Returns
    -------
    None
    """

    def __init__(
        self,
        list_of_grids,
        xlabel=None,
        ylabel=None,
        fn=None,
        title=None,
        fig_size=(10, 10),
    ):
        # Initialize the MosaicsGridData object with specified plot attributes
        # Inherits from Plotter and sets plot labels, title, and figure size.
        super().__init__(xlabel, ylabel, fn, title, fig_size)
        # load grid data from file
        self.grid_data_list = [np.loadtxt(file) for file in list_of_grids]

    def _plot_single(
        self, grid_data, axis, frame=None, title=None, v_min=None, v_max=None, **kwargs
    ):
        """
        Create a single heatmap plot.

        Parameters
        ----------
        grid_data : numpy.ndarray
            Grid data to be plotted.
        axis : matplotlib.axes.Axes
            The axis on which the heatmap will be plotted.
        frame : int, optional
            Frame parameter for the grid data.
        title : str, optional
            Title for the plot.
        v_min : float, optional
            Minimum value for color scaling.
        v_max : float, optional
            Maximum value for color scaling.
        **kwargs : dict
            Additional keyword arguments for the imshow function.

        Returns
        -------
        None
        """
        # Create heatmap using the grid data
        if frame is None:
            if v_min is not None:
                im = axis.imshow(
                    grid_data, origin="lower", vmin=v_min, vmax=v_max, **kwargs
                )
            else:
                im = axis.imshow(grid_data, origin="lower", **kwargs)
        else:
            if v_min is not None:
                im = axis.imshow(
                    grid_data[frame:-frame, frame:-frame],
                    origin="lower",
                    vmin=v_min,
                    vmax=v_max,
                    **kwargs
                )
            else:
                im = axis.imshow(
                    grid_data[frame:-frame, frame:-frame], origin="lower", **kwargs
                )

        axis.grid(False)

        # Create a colorbar of the same size as the plot
        divider = make_axes_locatable(axis)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax)
        # Set colorbar label and font size
        cbar.ax.tick_params(labelsize=12)

        # Remove labels and ticks from the axes
        axis.set_xticks([])
        axis.set_yticks([])
        axis.set_xticklabels([])
        axis.set_yticklabels([])

        # Add title to the plot
        if title is not None:
            axis.set_title(
                title,
                fontsize=12,
                weight="bold",
                pad=15,
                fontfamily="Arial Rounded MT Bold",
            )

    def create_plot(
        self,
        shape: tuple,
        frame=None,
        nan_to_zero=False,
        titles=None,
        share_limits=False,
        **kwargs
    ):
        """
        Create a mosaic of heatmap plots.

        Parameters
        ----------
        shape : tuple
            Shape of the mosaic as (rows, columns).
        frame : int, optional
            Frame parameter for the grid data.
        nan_to_zero : bool, optional
            Convert NaN values to zero.
        titles : list of str, optional
            Titles for individual subplots. If None, no titles are used.
        share_limits : bool, optional
            Share color scale limits across subplots.
        **kwargs : dict
            Additional keyword arguments for the _plot_single function.

        Returns
        -------
        None
        """

        if len(shape) != 2:
            raise ValueError("shape must be a tuple of length 2.")
        elif shape[0] * shape[1] != len(self.grid_data_list):
            raise ValueError(
                "The number of grids must be equal to the product of the shape."
            )
        elif len(self.grid_data_list) == 1:
            raise ValueError(
                "You only have one grid. We encourage you to use a single plot instead."
            )
        if titles is not None:
            if len(titles) != len(self.grid_data_list):
                raise ValueError(
                    "titles must have the same length as the number of grids."
                )
        if nan_to_zero:
            self.grid_data_list = [np.nan_to_num(arr) for arr in self.grid_data_list]
        if share_limits:
            v_min = 0
            v_max = 0
            for grid_data in self.grid_data_list:
                v_min = min(v_min, np.nanmin(grid_data))
                v_max = max(v_max, np.nanmax(grid_data))

        n_rows = shape[0]
        n_cols = shape[1]

        # Generate a figure and axis for the plot
        fig, axes = plt.subplots(
            n_rows,
            n_cols,
            figsize=self.fig_size,
        )

        if n_rows == 1 or n_cols == 1:
            for i, grid_data in enumerate(self.grid_data_list):
                if titles is None:
                    if share_limits:
                        self._plot_single(
                            grid_data,
                            axes[i],
                            frame=frame,
                            v_min=v_min,
                            v_max=v_max,
                            **kwargs
                        )
                    else:
                        self._plot_single(grid_data, axes[i], frame=frame, **kwargs)
                else:
                    if share_limits:
                        self._plot_single(
                            grid_data,
                            axes[i],
                            frame=frame,
                            title=titles[i],
                            v_min=v_min,
                            v_max=v_max,
                            **kwargs
                        )
                    else:
                        self._plot_single(
                            grid_data, axes[i], frame=frame, title=titles[i], **kwargs
                        )
        else:
            for i, grid_data in enumerate(self.grid_data_list):
                if titles is None:
                    if share_limits:
                        self._plot_single(
                            grid_data,
                            axes[i // n_cols, i % n_cols],
                            frame=frame,
                            v_min=v_min,
                            v_max=v_max,
                            **kwargs
                        )
                    else:
                        self._plot_single(
                            grid_data,
                            axes[i // n_cols, i % n_cols],
                            frame=frame,
                            **kwargs
                        )
                else:
                    if share_limits:
                        self._plot_single(
                            grid_data,
                            axes[i // n_cols, i % n_cols],
                            frame=frame,
                            title=titles[i],
                            v_min=v_min,
                            v_max=v_max,
                            **kwargs
                        )
                    else:
                        self._plot_single(
                            grid_data,
                            axes[i // n_cols, i % n_cols],
                            frame=frame,
                            title=titles[i] ** kwargs,
                        )

        # # Set default filename and title if not provided
        # if self.fn is None:
        #     self.fn = os.path.join(
        #         os.getcwd(), "mosaics_grid_data_{}.pdf".format(prop_label)
        #     )
        # if self.title is None:
        #     self.title = "Mosaics Grid Data - {}".format(prop_label)

        # # Add title to the plot
        # ax.set_title(
        #     self.title,
        #     fontsize=14,
        #     weight="bold",
        #     pad=15,
        #     fontfamily="Arial Rounded MT Bold",
        # )
        plt.tight_layout()



"""ProLint2: Lipid-Protein Interaction Analysis"""

# Add imports here
import os
from .core import Universe

# Handle versioneer
from ._version import get_versions

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



#!/usr/bin/env python

import numpy as np
from itertools import combinations
from prolint2.server.utils import calculate_contact_intervals

import os
import configparser

# Getting the config file
config = configparser.ConfigParser(allow_no_value=True)
config.read(os.path.join(os.path.abspath(os.path.dirname(__file__)), "../config.ini"))
parameters_config = config["Parameters"]

# def per_lipid_contacts(ts, lipids, frame_cutoff=10):
#     """
#     Given a list of lipid IDs, returns a dict with these lipid IDs
#     as keys, and values set to a dict containing the residues these lipids
#     interact with as keys and the corresponding number of contacts as values.
#     These contacts can be filtered using the `frame_cutoff` option.

#     TODO:
#     `frame_cutoff` should operate on a percentage of trajectory length.
#     """
#     results = {k: {} for k in lipids}
#     for k, v in ts.contacts.contact_frames.items():
#         if len(v) < frame_cutoff:
#             continue
#         # r, l = [int(x) for x in k.split(',')] # k used to be a string formatted as 'residue,lipid'
#         # k now is a tuple of (residue, lipid)
#         r, l = k
#         if l in lipids:
#             results[l][r] = len(v)
#     return results


def per_lipid_contacts(ts, lipids, frame_cutoff=10):
    results = {k: {} for k in lipids}
    for residue_id, lipid_dict in ts.contacts.contact_frames.items():
        for lipid_id, frames in lipid_dict.items():
            if len(frames) < frame_cutoff:
                continue
            if lipid_id in lipids:
                results[lipid_id][residue_id] = len(frames)
    return results


def sort_dict(d, cutoff=None):
    """
    Takes a dictionary as input, and sorts it according to values.
    The purpose of this function is to order residue:contacts dicts build
    by `per_lipid_contacts` function according to the number of contacts.

    The `cutoff` value returns only a subsection of this dict. The cutoff
    is used by `get_linked_nodes` to get a subsection of edges/links. This is
    important because visually we can not easily distinguish between link widths,
    so filtering here allows us to only show the top X links.
    """
    item_list = sorted(d.items(), key=lambda x: x[1], reverse=True)
    if cutoff:
        return dict(item_list[:cutoff])
    return dict(item_list)


def residue_pair_matching_contacts(res1_contacts, res2_contacts):
    """
    docstring: important improvement to the network app contacts
    """
    total_contacts = 0
    for s1, e1 in res1_contacts:
        for s2, e2 in res2_contacts:
            if not e1 < s2 and not s1 > e2:
                contacts = min(e1, e2) - max(s1, s2)
                total_contacts += contacts

    return total_contacts


def get_ordered_combinations(lipid_contacts):
    """
    This function turns residue:contact information into shared contacts. To speed up
    performance, shared contacts are define in an adhoc manner. It takes as input the
    returned dict by `per_lipid_contacts`. It then forms all possible size-2 combinations.

    For a lipid ID, all these residue-residue combinations are seen as shared contacts.
    That is, if a lipid interacts with residues 1, 2, and 3, then all possible size-2
    combination (1-2, 1-3, 2-3) are seen as shared contacts.

    These combinations are assigned as keys, with values corresponding to the sum
    of all matching shared contacts across all lipid IDs. This dict is then returned.
    This approach seems to work quite well in addition to being intuitive. One potential
    disadvantage, however, is that if a residue has very high contact count, then so will
    all combinations involving that residue.

    Another important point to consider is that since shared contacts are defined adhoc,
    an edge linking two nodes, cannot reliably be interpreted as geometric distance.
    Meaning, it does not mean that those two nodes/residues are connected by interactions
    with the same lipid.

    TODO:
    1.
    Is it possible to improve the last point, by counting only real shared contacts?
    The information is already contained in `contacts.contact_frames`.
    2.
    combinations call already returns an iterable. Simply by removing the
    subsequent call to list().
    """
    ordered_combinations = {}
    for lipid_id, lipid_vals in lipid_contacts.items():
        sorted_lipid_vals = sort_dict(lipid_vals)

        lipid_residues = list(sorted_lipid_vals.keys())
        for res1, res2 in list(combinations(lipid_residues, 2)):
            value = sum([sorted_lipid_vals[res1], sorted_lipid_vals[res2]])
            key = f"{res1},{res2}"

            if key in ordered_combinations:
                ordered_combinations[key] = ordered_combinations[key] + value
            else:
                ordered_combinations[key] = value

    return ordered_combinations


def shared_contacts(contacts, top_lipids, lipid_contact_frames, *args, **kwargs):
    """
    Aim: improve the shortcomings outlined in `get_ordered_combinations`.
    """
    lipid_shared_contacts = {}
    for lipid in top_lipids:
        contact_intervals = calculate_contact_intervals(
            contacts, lipid_contact_frames, lipid, *args, **kwargs
        )
        residue_contacts = {}
        for res1, res2 in combinations(contact_intervals.keys(), 2):
            pair_contacts = residue_pair_matching_contacts(
                contact_intervals[res1], contact_intervals[res2]
            )
            residue_contacts[f"{res1},{res2}"] = pair_contacts
        lipid_shared_contacts[lipid] = residue_contacts

    shared_contacts_all = {}
    for residue_pair_contacts in lipid_shared_contacts.values():
        for residue_pair, pair_contacts in residue_pair_contacts.items():
            if residue_pair in shared_contacts_all:
                shared_contacts_all[residue_pair] += pair_contacts
            else:
                shared_contacts_all[residue_pair] = pair_contacts

    # return sort_dict(shared_contacts_all)
    return lipid_shared_contacts, sort_dict(shared_contacts_all)


def get_linked_nodes(ordered_combinations, cutoff=100):
    """
    This function along with `get_node_list` help deal with amChart shortcomings
    related to ChordNonRibbon. We need to add hidden nodes to keep the chord
    architecture nice and matching the structure of the protein. Here we get
    the node ids that have data linked to them.
    """
    linked_nodes = []
    for residue_key in sort_dict(ordered_combinations, cutoff=cutoff).keys():
        res1, res2 = [int(res) for res in residue_key.split(",")]
        linked_nodes.extend([res1, res2])
    return np.unique(linked_nodes).tolist()


def get_node_list(n_residues, linked_nodes):
    """
    This function along with `get_linked_nodes` help deal with amChart shortcomings
    related to ChordNonRibbon. We need to add hidden nodes to keep the chord
    architecture nice and matching the structure of the protein. Here we get
    the node ids that do NOT have data linked to them.

    TODO:
    We need to have checks on the total number of nodes that can be visualized.
    Currently, we add 200 nodes on top of the already existing linked nodes.
    """

    hidden_nodes = 200

    node_list = np.linspace(0, n_residues, hidden_nodes, dtype=int)
    node_list = [int(x) for x in node_list if int(x) not in linked_nodes]
    nodes = sorted(node_list + linked_nodes)

    hidden_node_indices = [ix for (ix, x) in enumerate(nodes) if x not in linked_nodes]
    return nodes, hidden_node_indices


def get_chord_elements(ts, nodes, ordered_combinations, cutoff=500):
    """
    Prepares the input data so it can be read and understood by the amCharts.
    The output is a simple list of dicts. Each entry containing info for
    one link. We also add residue ID and residue name information which is
    important to link node data with other apps.
    """
    node_links = list(combinations(nodes, 2))
    position_node_links = [x for x in node_links if x[0] == 0]

    resnums = ts.query.residues.resnums
    resnames = ts.query.residues.resnames
    node_names = {x[0]: f"{x[0]} {x[1]}" for x in list(zip(resnums, resnames))}

    chord_elements = []
    for res1, res2 in position_node_links:
        chord_elements.append({"from": 0, "to": node_names[res2], "value": 0})

    contact_max = max(ordered_combinations.values())
    for k, v in sort_dict(ordered_combinations, cutoff=cutoff).items():
        res1, res2 = k.split(",")
        # This will break the nodes. We need to add the nodes we are skipping here
        # to the position_residue_indices
        # if abs(int(res1) - int(res2)) < 4:
        #     continue

        chord_elements.append(
            {
                "from": node_names[int(res1)],
                "to": node_names[int(res2)],
                "value": v,
                "valueWidth": float(v) / contact_max,
            }
        )

    return chord_elements


def contact_chord(ts, contacts, top_lipid_ids, lipid_contact_frames, cutoff=100):
    """
    We call all functions here. We return the chord elements (these are the data
    amCharts needs to render nodes and links), we also return information on which
    nodes to hide, and which nodes correspond to which lipid ID.

    TODO:
    `per_lipid_nodes` can be further refined.
    """
    # lipid_contacts = per_lipid_contacts(ts, top_lipid_ids)
    lipid_shared_contacts, ordered_combinations = shared_contacts(
        contacts,
        top_lipid_ids,
        lipid_contact_frames,
        residues_to_show=int(parameters_config["residues_to_show"]),
        intervals_to_filter_out=int(parameters_config["intervals_to_filter_out"]),
    )
    # ordered_combinations = get_ordered_combinations(lipid_contacts)
    linked_nodes = get_linked_nodes(ordered_combinations, cutoff=cutoff)
    nodes, hidden_node_indices = get_node_list(ts.query.n_residues, linked_nodes)
    chord_elements = get_chord_elements(ts, nodes, ordered_combinations, cutoff=cutoff)

    per_lipid_nodes = {}
    for lipid, pair_contacts in lipid_shared_contacts.items():
        lipid_nodes = get_linked_nodes(pair_contacts, cutoff=None)
        all_lipid_nodes = [x for x in lipid_nodes if x in linked_nodes]
        if all_lipid_nodes:
            per_lipid_nodes[lipid] = all_lipid_nodes

    return chord_elements, hidden_node_indices, per_lipid_nodes



from prolint2.plotting.plotting import (
    Plotter,
    PointDistribution,
    Radar,
    DensityMap,
    DurationGantt,
    LogoResidues,
    InteractionHeatMap,
    MetricsComparison,
    SharedContacts,
    TwoPointDistanceEvolution,
    MosaicsGridData,
)

__all__ = [
    "Plotter",
    "PointDistribution",
    "Radar",
    "DensityMap",
    "DurationGantt",
    "LogoResidues",
    "InteractionHeatMap",
    "MetricsComparison",
    "SharedContacts",
    "TwoPointDistanceEvolution",
    "MosaicsGridData",
]



from abc import ABC, abstractmethod
from MDAnalysis.analysis.base import AnalysisBase


class ContactComputerBase(AnalysisBase, ABC):
    # @abstractmethod
    # def _compute_pairs(self):
    #     pass

    # @abstractmethod
    # def _compute(self):
    #     pass

    def __add__(self, other):
        pass

    def intersection(self, other):
        pass

    def union(self, other):
        pass



import numpy as np


def fast_unique_comparison(residue_ids, lipid_ids, lipid_names):
    """
    Get the unique combinations of residue and lipid ids. Vectorized implementation.

    Parameters
    ----------
    residue_ids : np.ndarray
        Array of residue ids.
    lipid_ids : np.ndarray
        Array of lipid ids.
    lipid_names : np.ndarray
        Array of lipid names.

    Returns
    -------
    np.ndarray: Array of unique combinations of residue and lipid ids.
    """
    # Combine the arrays into a single 2D array
    combined_array = np.stack((residue_ids, lipid_ids), axis=-1)

    # Get lexicographically sorted indices
    lex_sorted_indices = np.lexsort((combined_array[:, 1], combined_array[:, 0]))

    # Sort the combined array by the sorted indices
    sorted_array = combined_array[lex_sorted_indices]

    # Calculate row-wise differences between consecutive sorted rows
    row_diffs = np.diff(sorted_array, axis=0)

    # Find the indices where the differences are non-zero
    unique_indices = np.where(np.any(row_diffs != 0, axis=1))[0]

    # Add the first index (0) to unique_indices, as it's always unique
    unique_indices = np.concatenate(([0], unique_indices + 1))

    # Extract the unique rows using the indices
    unique_array = sorted_array[unique_indices]

    # Split the unique rows back into residue_ids and lipid_ids
    unique_residue_ids, unique_lipid_ids = unique_array[:, 0], unique_array[:, 1]

    # Extract the corresponding lipid_names using the sorted indices
    sorted_lipid_names = lipid_names[lex_sorted_indices]
    unique_lipid_names = sorted_lipid_names[unique_indices]

    return unique_residue_ids, unique_lipid_ids, unique_lipid_names



from enum import Enum


# pylint: disable=invalid-name
class UnitConversionFactor(Enum):
    """Conversion factors for time units."""

    fs = 1e-15
    ps = 1e-12
    ns = 1e-9
    us = 1e-6
    ms = 1e-3
    s = 1.0


DEFAULT_SIM_PARAMS = {
    "units": "us",
    "normalizer": "actual_time",
    "unit_conversion_factor": UnitConversionFactor.ps.value
    / UnitConversionFactor.us.value,
    "norm_factor": 1,
}
